function [source test scores] = run_cvpr12(c)
% Run an experiment as described in the configuration c.
% Intermediate results are saved and reloaded on subsequent runs.
% Thus, not recomputed. To use date from ealier experiments,
% copy the relevant data to the folder of this experiment.
%
% arguments:
%   c = configuration struct
%     see st_config.m
%
% returns:
%   train = source for transfer (built from training data)
%   test = test data and prediction
%   scores = scores

if ~file_exists(c.folder)
	file_mkdir(c.folder);
end;

source = cached('train_source',@getSource,c);
test = transfer(c,source);
scores = cached('test_scores',@getScores,test);

% report
fprintf('average IoU = %02d%%\n',round(100*mean(scores.ious)));
fprintf('average accuracy = %.1f%%\n',100*mean(scores.accs));
fprintf('precision/recall F1 measure = %.1f%%\n',100*scores.f1);


function result = cached(name,func,varargin)

f = sprintf('%s/%s.mat',c.folder,name);

if file_exists(f)
	result = load_value(f);
else
	result = func(varargin{:});
	save_value(f,result);
end;

end


function source = getSource(c)

images = load_value(c.train_images);
segmentations = load_value(c.train_segmentations);
windows = cached('train_windows',@st_windows,images,c.windows);
if ischar(c.itq)
	itq = load_value(c.itq);
	features = cached('train_features',@st_features_itq,images,windows,@st_features_gist,{},itq);
else
	assert(~c.itq);
	features = cached('train_features',@st_features_gist,images,windows);
end;
clear images;
source = st_source(c.masksize,segmentations,windows,features);
clear segmentations windows features;

end


function test = transfer(c,source)

images = load_value(c.test_images);
segmentations = load_value(c.test_segmentations);
windows = cached('test_windows',@st_windows,images,c.windows);
if ischar(c.itq)
	itq = load_value(c.itq);
	features = cached('test_features',@st_features_itq,images,windows,@st_features_gist,{},itq);
else
	assert(~c.itq);
	features = cached('test_features',@st_features_gist,images,windows);
end;
if ischar(c.weights)
	weights = load_value(c.weights);
else
	weights = c.weights;
end;
masks = cached('test_masks',@st_transfer,source,c.masksize,windows,features,c.distance,weights);
clear windows features;
prediction = cached('test_prediction',@st_segment,images,masks,c.threshold,c.iterations);

collectStruct test images segmentations masks prediction;

end


function scores = getScores(test)

ious = st_score_iou(test.segmentations,test.prediction);
accs = st_score_acc(test.segmentations,test.prediction);
[~,~,f1] = st_score_pr(test.segmentations,test.prediction);

collectStruct scores ious accs f1;

end

end

