function c = st_config(name)
% Returns predefined experiment configurations.
%
% arguments
%   name = name of experiment configuration (default=voc10)
%     available names:
%       voc10 = pascal VOC 2010 segmentation challenge dataset
%       horses = weizmann horses dataset
%       graz_bikes = bikes subset of the Graz-02 dataset
%       graz_cars = cars subset of the Graz-02 dataset
%       graz_people = people subset of the Graz-02 dataset
%
% returns
%   c = experiment configuration struct
%     folder = experiment folder (holds intermediate data)
%     train_images = mat file with training images
%     train_segmentations = mat file with training ground truth
%     windows = number of windows per image
%     masksize = size of masks for transfer
%     test_images = mat file with test images
%     test_segmentations = mat file with test ground truth (for scoring)
%     itq = false for no itq, or mat file with itq configuration
%     distance = handle to distance function to be used (distance_*)
%     weights = weights for neighbors, or mat file with weights
%     threshold = threshold on the location prior for the first grabcut iteration
%     iterations = number of iterations for GrabCut

if ~exist('name','var')
	name = 'voc10';
end;

switch name
	case 'voc10'
		c.folder = 'data/xps/voc10';
		c.train_images = 'data/sets/voc10/train_images.mat';
		c.train_segmentations = 'data/sets/voc10/train_segmentations.mat';
		c.windows = 100;
		c.masksize = 50;
		c.test_images = 'data/sets/voc10/test_images.mat';
		c.test_segmentations = 'data/sets/voc10/test_segmentations.mat';
		c.itq = false;
		c.distance = @distance_L2;
		c.weights = ones(1,100)/100;
		c.threshold = 0.2;
		c.iterations = 20;
	case 'voc10_itq'
		c.folder = 'data/xps/voc10_itq';
		c.train_images = 'data/sets/voc10/train_images.mat';
		c.train_segmentations = 'data/sets/voc10/train_segmentations.mat';
		c.windows = 100;
		c.masksize = 50;
		c.test_images = 'data/sets/voc10/test_images.mat';
		c.test_segmentations = 'data/sets/voc10/test_segmentations.mat';
		c.itq = 'data/sets/voc10/train_itq.mat';
		c.distance = @distance_hamming;
		c.weights = ones(1,100)/100;
		c.threshold = 0.2;
		c.iterations = 20;
	case 'horses'
		c.folder = 'data/xps/horses';
		c.train_images = 'data/sets/horses/train_images.mat';
		c.train_segmentations = 'data/sets/horses/train_segmentations.mat';
		c.windows = 100;
		c.masksize = 50;
		c.test_images = 'data/sets/horses/test_images.mat';
		c.test_segmentations = 'data/sets/horses/test_segmentations.mat';
		c.itq = false;
		c.distance = @distance_L2;
		c.weights = ones(1,100)/100;
		c.threshold = 0.45;
		c.iterations = 20;
	case 'graz_bikes'
		c.folder = 'data/xps/graz/bikes';
		c.train_images = 'data/sets/graz/bikes/train_images.mat';
		c.train_segmentations = 'data/sets/graz/bikes/train_segmentations.mat';
		c.windows = 100;
		c.masksize = 50;
		c.test_images = 'data/sets/graz/bikes/test_images.mat';
		c.test_segmentations = 'data/sets/graz/bikes/test_segmentations.mat';
		c.itq = false;
		c.distance = @distance_L2;
		c.weights = ones(1,10)/10;
		c.threshold = 0.75;
		c.iterations = 20;
	case 'graz_cars'
		c.folder = 'data/xps/graz/cars';
		c.train_images = 'data/sets/graz/cars/train_images.mat';
		c.train_segmentations = 'data/sets/graz/cars/train_segmentations.mat';
		c.windows = 100;
		c.masksize = 50;
		c.test_images = 'data/sets/graz/cars/test_images.mat';
		c.test_segmentations = 'data/sets/graz/cars/test_segmentations.mat';
		c.itq = false;
		c.distance = @distance_L2;
		c.weights = ones(1,10)/10;
		c.threshold = 0.40;
		c.iterations = 20;
	case 'graz_people'
		c.folder = 'data/xps/graz/people';
		c.train_images = 'data/sets/graz/people/train_images.mat';
		c.train_segmentations = 'data/sets/graz/people/train_segmentations.mat';
		c.windows = 100;
		c.masksize = 50;
		c.test_images = 'data/sets/graz/people/test_images.mat';
		c.test_segmentations = 'data/sets/graz/people/test_segmentations.mat';
		c.itq = false;
		c.distance = @distance_L2;
		c.weights = ones(1,10)/10;
		c.threshold = 0.20;
		c.iterations = 20;
	otherwise
		assert(false,'unknown configuration name %s',name);
end;

