function gist = st_features_gist(images,windows)
% Computes the gist descriptor for windows in images.
%
% arguments
%   images = rgb uint8 image, or cell of ~
%   windows = array of windows, or cell of ~
%
% returns
%   gist = array of rows of features per window, or cell of ~
%

if iscell(images)
	% TODO check how much overhead for filter generation, generate and static argument? or batch?
	gist = map('c','par',@st_features_gist,'cc',{images,windows});
else
	gist = compute({images},{windows});
	gist = gist{1};
end;


function gist = compute(images,windows)
% Adapted gist computation from: http://people.csail.mit.edu/torralba/code/spatialenvelope/
% Modeling the shape of the scene: a holistic representation of the spatial envelope
% Aude Oliva, Antonio Torralba
% International Journal of Computer Vision, Vol. 42(3): 145-175, 2001.
%
% Uses the same call syntax as the main function,
% but doesn't allow for a single input image.

% settings
param.boundaryExtension = 32; % number of pixels to pad
param.imageSize = 128;
param.orientationsPerScale = [8 8 8 8];
param.numberBlocks = 4;
param.fc_prefilt = 4;
param.G = createGabor(param.orientationsPerScale,param.imageSize+2*param.boundaryExtension);

nImages = length(images);
D = size(param.G,3)*param.numberBlocks^2;

gist = cell(nImages,1);

for i = 1:nImages

	nWindows = size(windows{i},1);
	gist{i} = zeros(nWindows,3*D,'single');

	H = size(images{i},1);
	W = size(images{i},2);

	for w = 1:nWindows
		
		x1 = round(1+windows{i}(w,1)*(W-1));
		y1 = round(1+windows{i}(w,2)*(H-1));
		x2 = round(1+windows{i}(w,3)*(W-1));
		y2 = round(1+windows{i}(w,4)*(H-1));

		% extract window
		img = images{i}(y1:y2,x1:x2,:);

		% grayscale
%    img = single(mean(img,3));
		% color channels
		img1 = single(img(:,:,1));
		img2 = single(img(:,:,2));
		img3 = single(img(:,:,3));

    % resize and crop image to make it square
		% TODO that's now what matthieu was doing, really crop and scale?
%    img = imresizecrop(img,param.imageSize,'bilinear');
    %img = imresize(img, param.imageSize, 'bilinear'); %jhhays
    img1 = imresizecrop(img1,param.imageSize,'bilinear');
    img2 = imresizecrop(img2,param.imageSize,'bilinear');
    img3 = imresizecrop(img3,param.imageSize,'bilinear');

    % scale intensities to be in the range [0 255]
%    img = img-min(img(:));
%    img = 255*img/max(img(:));
    img1 = img1-min(img1(:));
    img1 = 255*img1/max(img1(:));
    img2 = img2-min(img2(:));
    img2 = 255*img2/max(img2(:));
    img3 = img3-min(img3(:));
    img3 = 255*img3/max(img3(:));
    
    % prefiltering: local contrast scaling
%    output = prefilt(img,param.fc_prefilt);
    output1 = prefilt(img1,param.fc_prefilt);
    output2 = prefilt(img2,param.fc_prefilt);
    output3 = prefilt(img3,param.fc_prefilt);

    % get gist:
%    gist{i}(w,:) = gistGabor(output,param);
		gist{i}(w,:) = [ gistGabor(output1,param)' gistGabor(output2,param)' gistGabor(output3,param)' ];

	end;
end


function output = prefilt(img, fc)
% ima = prefilt(img, fc);
% fc  = 4 (default)
% 
% Input images are double in the range [0, 255];
% You can also input a block of images [ncols nrows 3 Nimages]
%
% For color images, normalization is done by dividing by the local
% luminance variance.

if nargin == 1
    fc = 4; % 4 cycles/image
end

w = 5;
s1 = fc/sqrt(log(2));

% Pad images to reduce boundary artifacts
img = log(img+1);
img = padarray(img, [w w], 'symmetric');
[sn, sm, c, N] = size(img);
n = max([sn sm]);
n = n + mod(n,2);
img = padarray(img, [n-sn n-sm], 'symmetric','post');

% Filter
[fx, fy] = meshgrid(-n/2:n/2-1);
gf = fftshift(exp(-(fx.^2+fy.^2)/(s1^2)));
gf = repmat(gf, [1 1 c N]);

% Whitening
output = img - real(ifft2(fft2(img).*gf));
clear img

% Local contrast normalization
localstd = repmat(sqrt(abs(ifft2(fft2(mean(output,3).^2).*gf(:,:,1,:)))), [1 1 c 1]); 
output = output./(.2+localstd);

% Crop output to have same size than the input
output = output(w+1:sn-w, w+1:sm-w,:,:);


function g = gistGabor(img, param)
% 
% Input:
%   img = input image (it can be a block: [nrows, ncols, c, Nimages])
%   param.w = number of windows (w*w)
%   param.G = precomputed transfer functions
%
% Output:
%   g: are the global features = [Nfeatures Nimages], 
%                    Nfeatures = w*w*Nfilters*c

img = single(img);

w = param.numberBlocks;
G = param.G;
be = param.boundaryExtension;

if ndims(img)==2
    c = 1; 
    N = 1;
    [nrows ncols c] = size(img);
end
if ndims(img)==3
    [nrows ncols c] = size(img);
    N = c;
end
if ndims(img)==4
    [nrows ncols c N] = size(img);
    img = reshape(img, [nrows ncols c*N]);
    N = c*N;
end

[ny nx Nfilters] = size(G);
W = w*w;
g = zeros([W*Nfilters N]);

% pad image
img = padarray(img, [be be], 'symmetric');

img = single(fft2(img)); 
k=0;
for n = 1:Nfilters
    ig = abs(ifft2(img.*repmat(G(:,:,n), [1 1 N]))); 
    ig = ig(be+1:ny-be, be+1:nx-be, :);
    
    v = downN(ig, w);
    g(k+1:k+W,:) = reshape(v, [W N]);
    k = k + W;
    drawnow
end

if c == 3
    % If the input was a color image, then reshape 'g' so that one column
    % is one images output:
    g = reshape(g, [size(g,1)*3 size(g,2)/3]);
end


function y=downN(x, N)
% 
% averaging over non-overlapping square image blocks
%
% Input
%   x = [nrows ncols nchanels]
% Output
%   y = [N N nchanels]

nx = fix(linspace(0,size(x,1),N+1));
ny = fix(linspace(0,size(x,2),N+1));
y  = zeros(N, N, size(x,3));
for xx=1:N
  for yy=1:N
    v=mean(mean(x(nx(xx)+1:nx(xx+1), ny(yy)+1:ny(yy+1),:),1),2);
    y(xx,yy,:)=v(:);
  end
end


function G = createGabor(or, n)
%
% G = createGabor(numberOfOrientationsPerScale, n);
%
% Precomputes filter transfer functions. All computations are done on the
% Fourier domain. 
%
% If you call this function without output arguments it will show the
% tiling of the Fourier domain.
%
% Input
%     numberOfOrientationsPerScale = vector that contains the number of
%                                orientations at each scale (from HF to BF)
%     n = imagesize = [nrows ncols] 
%
% output
%     G = transfer functions for a jet of gabor filters


Nscales = length(or);
Nfilters = sum(or);

if length(n) == 1
    n = [n(1) n(1)];
end

l=0;
for i=1:Nscales
    for j=1:or(i)
        l=l+1;
        param(l,:)=[.35 .3/(1.85^(i-1)) 16*or(i)^2/32^2 pi/(or(i))*(j-1)];
    end
end

% Frequencies:
%[fx, fy] = meshgrid(-n/2:n/2-1);
[fx, fy] = meshgrid(-n(2)/2:n(2)/2-1, -n(1)/2:n(1)/2-1);
fr = fftshift(sqrt(fx.^2+fy.^2));
t = fftshift(angle(fx+sqrt(-1)*fy));

% Transfer functions:
G=zeros([n(1) n(2) Nfilters]);
for i=1:Nfilters
    tr=t+param(i,4); 
    tr=tr+2*pi*(tr<-pi)-2*pi*(tr>pi);

    G(:,:,i)=exp(-10*param(i,1)*(fr/n(2)/param(i,2)-1).^2-2*param(i,3)*pi*tr.^2);
end


if nargout == 0
    figure
    for i=1:Nfilters
        contour(fx, fy, fftshift(G(:,:,i)),[1 .7 .6],'r');
        hold on
    end
    axis('on')
    axis('equal')
    axis([-n(2)/2 n(2)/2 -n(1)/2 n(1)/2])
    axis('ij')
    xlabel('f_x (cycles per image)')
    ylabel('f_y (cycles per image)')
    grid on
end


function img = imresizecrop(img, M, METHOD)
%
% img = imresizecrop(img, M, METHOD);
%
% Output an image of size M(1) x M(2).

if nargin < 3
    METHOD = 'bilinear';
end

if length(M) == 1
    M = [M(1) M(1)];
end

scaling = max([M(1)/size(img,1) M(2)/size(img,2)]);

%scaling = M/min([size(img,1) size(img,2)]);

newsize = round([size(img,1) size(img,2)]*scaling);
img = imresize(img, newsize, METHOD);

[nr nc cc] = size(img);

sr = floor((nr-M(1))/2);
sc = floor((nc-M(2))/2);

img = img(sr+1:sr+M(1), sc+1:sc+M(2),:);

