function mask = st_transfer(source,outsize,windows,features,distance,weights)
% Transfers segmentations from source(s) to images.
%
% arguments:
%   source = one source or cell of sources
%   outsize = size for transfer output
%     [H W] = output is HxW
%     S = output is SxS
%   windows = cell of windows or cell of cell of windows
%   features = cell of features or cell of cell of features
%   distance = function handle to distance_*
%     distance to be used for the features
%   weights = weights for neighbors
%     1d array of weights
%     should sum to one
%
% returns:
%   mask = logical mask or cell of logical mask
%     the size is determined by outsize

if iscell(source)
	source = st_source_join(source);
end;

if length(outsize)==1
	outsize = [ outsize outsize ];
else
	assert(length(outsize)==2);
end;

if iscell(windows)
	mask = map('c','par',@compute,'--cc--',{source,outsize,windows,features,distance,weights});
else
	mask = compute(source,outsize,windows,features,distance,weights);
end;


function mask = compute(source,outsize,windows,features,distance,weights)
% Compute transfer for one image only.

M = source.M;

K = length(weights);
weights = reshape(weights,1,1,K);
weights = repmat(weights,[M,M,1]);

n = size(windows,1);
assert(n==size(features,1));
assert(size(features,2)==source.D);
assert(K>0);
assert(K<=source.N);

% transform normalized windows to outsize
H = outsize(1);
W = outsize(2);
windows(:,1) = round(windows(:,1)*(W-1)+1);
windows(:,2) = round(windows(:,2)*(H-1)+1);
windows(:,3) = round(windows(:,3)*(W-1)+1);
windows(:,4) = round(windows(:,4)*(H-1)+1);

mask = zeros(H,W);

for w = 1:n
	
	d = distance(features(w,:),source.features);
	[~,idx] = sort(d,'ascend');
	idx = idx(1:K);

	t = sum( source.masks(:,:,idx) .* weights , 3 );
	
	y = windows(w,2):windows(w,4);
	x = windows(w,1):windows(w,3);

	v = round(linspace(1,M,length(y)));
	u = round(linspace(1,M,length(x)));

	mask(y,x) = mask(y,x) + t(v,u);

end;

m = max(mask(:));
if m>0
	mask = mask ./ m;
end;
